@@ -1,15 +1,12 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2019-2020 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2019-2023 -- leonerd@leonerd.org.uk
 
-package Future::IO;
+package Future::IO 0.12;
 
-use strict;
+use v5.14;
 use warnings;
-use 5.010;  # //
-
-our $VERSION = '0.10';
 
 use Carp;
 
@@ -20,6 +17,9 @@ my @writers;
 
 our $IMPL;
 
+our $MAX_READLEN = 8192;
+our $MAX_WRITELEN = 8192;
+
 =head1 NAME
 
 C<Future::IO> - Future-returning IO methods
@@ -95,14 +95,34 @@ A testing module which does this is provided by L<Test::Future::IO>.
 
 =cut
 
+=head2 accept
+
+   $socketfh = await Future::IO->accept( $fh );
+
+I<Since version 0.11.>
+
+Returns a L<Future> that will become done when a new connection has been
+accepted on the given filehandle, which should represent a listen-mode socket.
+The returned future will yield the newly-accepted client socket filehandle.
+
+=cut
+
+sub accept
+{
+   shift;
+   my ( $fh ) = @_;
+
+   return ( $IMPL //= "Future::IO::_DefaultImpl" )->accept( $fh );
+}
+
 =head2 alarm
 
-   $f = Future::IO->alarm( $epoch )
+   await Future::IO->alarm( $epoch );
 
 I<Since version 0.08.>
 
-Returns a L<Future> that will become at a fixed point in the future, given as
-an epoch timestamp (such as returned by C<time()>). This value may be
+Returns a L<Future> that will become done at a fixed point in the future,
+given as an epoch timestamp (such as returned by C<time()>). This value may be
 fractional.
 
 =cut
@@ -122,9 +142,28 @@ sub alarm
    }
 }
 
+=head2 connect
+
+   await Future::IO->connect( $fh, $name );
+
+I<Since version 0.11.>
+
+Returns a L<Future> that will become done when a C<connect()> has succeeded on
+the given filehandle to the given sockname address.
+
+=cut
+
+sub connect
+{
+   shift;
+   my ( $fh, $name ) = @_;
+
+   return ( $IMPL //= "Future::IO::_DefaultImpl" )->connect( $fh, $name );
+}
+
 =head2 sleep
 
-   $f = Future::IO->sleep( $secs )
+   await Future::IO->sleep( $secs );
 
 Returns a L<Future> that will become done a fixed delay from now, given in
 seconds. This value may be fractional.
@@ -141,9 +180,7 @@ sub sleep
 
 =head2 sysread
 
-   $f = Future::IO->sysread( $fh, $length )
-
-      $bytes = $f->get
+   $bytes = await Future::IO->sysread( $fh, $length );
 
 Returns a L<Future> that will become done when at least one byte can be read
 from the given filehandle. It may return up to C<$length> bytes. On EOF, the
@@ -166,9 +203,7 @@ sub sysread
 
 =head2 sysread_exactly
 
-   $f = Future::IO->sysread_exactly( $fh, $length )
-
-      $bytes = $f->get
+   $bytes = await Future::IO->sysread_exactly( $fh, $length );
 
 I<Since version 0.03.>
 
@@ -194,29 +229,62 @@ sub sysread_exactly
       return $IMPL->$code( $fh, $length );
    }
 
-   return _sysread_into_buffer( $IMPL, $fh, $length, '' );
+   return _sysread_into_buffer( $IMPL, $fh, $length, \(my $buffer = '') );
 }
 
 sub _sysread_into_buffer
 {
-   my ( $IMPL, $fh, $length, $bytes ) = @_;
+   my ( $IMPL, $fh, $length, $bufref ) = @_;
 
-   $IMPL->sysread( $fh, $length - length $bytes )->then( sub {
+   $IMPL->sysread( $fh, $length - length $$bufref )->then( sub {
       my ( $more ) = @_;
       return Future->done() if !defined $more; # EOF
 
-      $bytes .= $more;
+      $$bufref .= $more;
 
-      return Future->done( $bytes ) if length $bytes >= $length;
-      return _sysread_into_buffer( $IMPL, $fh, $length, $bytes );
+      return Future->done( $$bufref ) if length $$bufref >= $length;
+      return _sysread_into_buffer( $IMPL, $fh, $length, $bufref );
+   });
+}
+
+=head2 sysread_until_eof
+
+   $f = Future::IO->sysread_until_eof( $fh )
+
+I<Since version 0.12.>
+
+Returns a L<Future> that will become done when the given filehandle reaches
+the EOF condition. The returned future will yield all of the bytes read up
+until that point.
+
+=cut
+
+sub sysread_until_eof
+{
+   shift;
+   my ( $fh ) = @_;
+
+   $IMPL //= "Future::IO::_DefaultImpl";
+
+   return _sysread_until_eof( $IMPL, $fh, \(my $buffer = '') );
+}
+
+sub _sysread_until_eof
+{
+   my ( $IMPL, $fh, $bufref ) = @_;
+
+   $IMPL->sysread( $fh, $MAX_READLEN )->then( sub {
+      my ( $more ) = @_;
+      return Future->done( $$bufref ) if !defined $more;
+
+      $$bufref .= $more;
+      return _sysread_until_eof( $IMPL, $fh, $bufref );
    });
 }
 
 =head2 syswrite
 
-   $f = Future::IO->syswrite( $fh, $bytes )
-
-      $written_len = $f->get
+   $written_len = await Future::IO->syswrite( $fh, $bytes );
 
 I<Since version 0.04.>
 
@@ -240,9 +308,7 @@ sub syswrite
 
 =head2 syswrite_exactly
 
-   $f = Future::IO->syswrite_exactly( $fh, $bytes )
-
-      $written_len = $f->get
+   $written_len = await Future::IO->syswrite_exactly( $fh, $bytes );
 
 I<Since version 0.04.>
 
@@ -266,27 +332,25 @@ sub syswrite_exactly
       return $IMPL->$code( $fh, $bytes );
    }
 
-   return _syswrite_from_buffer( $IMPL, $fh, $bytes, length $bytes );
+   return _syswrite_from_buffer( $IMPL, $fh, \$bytes, length $bytes );
 }
 
 sub _syswrite_from_buffer
 {
-   my ( $IMPL, $fh, $bytes, $len ) = @_;
+   my ( $IMPL, $fh, $bufref, $len ) = @_;
 
-   $IMPL->syswrite( $fh, $bytes )->then( sub {
+   $IMPL->syswrite( $fh, substr $$bufref, 0, $MAX_WRITELEN )->then( sub {
       my ( $written_len ) = @_;
-      substr $bytes, 0, $written_len, "";
+      substr $$bufref, 0, $written_len, "";
 
-      return Future->done( $len ) if !length $bytes;
-      return _syswrite_from_buffer( $IMPL, $fh, $bytes, $len );
+      return Future->done( $len ) if !length $$bufref;
+      return _syswrite_from_buffer( $IMPL, $fh, $bufref, $len );
    });
 }
 
 =head2 waitpid
 
-   $f = Future::IO->waitpid( $pid )
-
-      $wstatus = $f->get
+   $wstatus = await Future::IO->waitpid( $pid );
 
 I<Since version 0.09.>
 
@@ -314,7 +378,7 @@ sub waitpid
 
 =head2 override_impl
 
-   Future::IO->override_impl( $impl )
+   Future::IO->override_impl( $impl );
 
 Sets a new implementation for C<Future::IO>, replacing the minimal default
 internal implementation. This can either be a package name or an object
@@ -344,19 +408,37 @@ sub override_impl
    ( $IMPL ) = @_;
 }
 
+=head2 HAVE_MULTIPLE_FILEHANDLES
+
+   $has = Future::IO->HAVE_MULTIPLE_FILEHANDLES;
+
+I<Since version 0.11.>
+
+Returns true if the underlying IO implementation actually supports multiple
+filehandles. Most real support modules will return true here, but this returns
+false for the internal minimal implementation.
+
+=cut
+
+sub HAVE_MULTIPLE_FILEHANDLES
+{
+   return ( $IMPL //= "Future::IO::_DefaultImpl" )->HAVE_MULTIPLE_FILEHANDLES;
+}
+
 package
    Future::IO::_DefaultImpl;
-use base qw( Future );
+use base qw( Future::IO::ImplBase );
 use Carp;
 
-use Errno qw( EAGAIN EWOULDBLOCK );
 use Struct::Dumb qw( readonly_struct );
 use Time::HiRes qw( time );
 
 readonly_struct Alarm => [qw( time f )];
 
-readonly_struct Reader => [qw( fh length f )];
-readonly_struct Writer => [qw( fh bytes f )];
+readonly_struct Reader => [qw( fh f )];
+readonly_struct Writer => [qw( fh f )];
+
+use constant HAVE_MULTIPLE_FILEHANDLES => 0;
 
 sub alarm
 {
@@ -370,50 +452,50 @@ sub sleep
    return $class->_done_at( time() + shift );
 }
 
-sub sysread
+sub ready_for_read
 {
    my $class = shift;
-   my ( $fh, $length ) = @_;
+   my ( $fh ) = @_;
 
    croak "This implementation can only cope with a single pending filehandle in ->syread"
       if @readers and $readers[-1]->fh != $fh;
 
-   my $self = $class->new;
-   push @readers, Reader( $fh, $length, $self );
+   my $f = Future::IO::_DefaultImpl::F->new;
+   push @readers, Reader( $fh, $f );
 
-   $self->on_cancel( sub {
-      my $self = shift;
+   $f->on_cancel( sub {
+      my $f = shift;
 
       my $idx = 0;
-      $idx++ while $idx < @readers and $readers[$idx]->f != $self;
+      $idx++ while $idx < @readers and $readers[$idx]->f != $f;
 
       splice @readers, $idx, 1, ();
    });
 
-   return $self;
+   return $f;
 }
 
-sub syswrite
+sub ready_for_write
 {
    my $class = shift;
-   my ( $fh, $bytes ) = @_;
+   my ( $fh ) = @_;
 
    croak "This implementation can only cope with a single pending filehandle in ->syswrite"
       if @writers and $writers[-1]->fh != $fh;
 
-   my $self = $class->new;
-   push @writers, Writer( $fh, $bytes, $self );
+   my $f = Future::IO::_DefaultImpl::F->new;
+   push @writers, Writer( $fh, $f );
 
-   $self->on_cancel( sub {
-      my $self = shift;
+   $f->on_cancel( sub {
+      my $f = shift;
 
       my $idx = 0;
-      $idx++ while $idx < @writers and $writers[$idx]->f != $self;
+      $idx++ while $idx < @writers and $writers[$idx]->f != $f;
 
       splice @writers, $idx, 1, ();
    });
 
-   return $self;
+   return $f;
 }
 
 sub waitpid
@@ -423,27 +505,34 @@ sub waitpid
 
 sub _done_at
 {
-   my $self = shift->new;
+   shift;
    my ( $time ) = @_;
 
+   my $f = Future::IO::_DefaultImpl::F->new;
+
    # TODO: Binary search
    my $idx = 0;
    $idx++ while $idx < @alarms and $alarms[$idx]->time < $time;
 
-   splice @alarms, $idx, 0, Alarm( $time, $self );
+   splice @alarms, $idx, 0, Alarm( $time, $f );
 
-   $self->on_cancel( sub {
+   $f->on_cancel( sub {
       my $self = shift;
 
       my $idx = 0;
-      $idx++ while $idx < @alarms and $alarms[$idx]->f != $self;
+      $idx++ while $idx < @alarms and $alarms[$idx]->f != $f;
 
       splice @alarms, $idx, 1, ();
    } );
 
-   return $self;
+   return $f;
 }
 
+package # hide
+   Future::IO::_DefaultImpl::F;
+use base qw( Future );
+use Time::HiRes qw( time );
+
 sub _await_once
 {
    die "Cowardly refusing to sit idle and do nothing" unless @alarms || @readers || @writers;
@@ -479,43 +568,10 @@ redo_select:
    }
 
    if( $rready ) {
-      my $r = $readers[0];
-
-      my $len = $r->fh->sysread( my $buf, $r->length );
-      if( $len ) {
-         shift @readers;
-         $r->f->done( $buf );
-      }
-      elsif( defined $len ) {
-         # EOF
-         shift @readers;
-         $r->f->done();
-      }
-      elsif( $! == EAGAIN or $! == EWOULDBLOCK ) {
-         $do_select = 1, goto redo_select if !$do_select;
-         # ignore it
-      }
-      else {
-         shift @readers;
-         $r->f->fail( "sysread: $!\n", sysread => $r->fh, $! );
-      }
+      ( shift @readers )->f->done;
    }
    if( $wready ) {
-      my $w = $writers[0];
-
-      my $len = $w->fh->syswrite( $w->bytes );
-      if( $len ) {
-         shift @writers;
-         $w->f->done( $len );
-      }
-      elsif( $! == EAGAIN or $! == EWOULDBLOCK ) {
-         $do_select = 1, goto redo_select if !$do_select;
-         # ignore it
-      }
-      else {
-         shift @writers;
-         $w->f->fail( "syswrite: $!\n", syswrite => $w->fh, $! );
-      }
+      ( shift @writers )->f->done;
    }
 
    my $now = time();
@@ -526,7 +582,7 @@ redo_select:
 
 =head2 await
 
-   $f = $f->await
+   $f = $f->await;
 
 I<Since version 0.07.>
 
@@ -542,9 +598,6 @@ sub await
    return $self;
 }
 
-# TODO: Consider implementing the _exactly variants of sysread/syswrite for
-#   efficiency
-
 =head1 THE C<$IMPL> VARIABLE
 
 I<Since version 0.02.>
