@@ -3,8 +3,7 @@
 #
 #  (C) Paul Evans, 2023 -- leonerd@leonerd.org.uk
 
-package Future::IO::System;
-our $VERSION = '0.17_003';
+package Future::IO::System 0.12;
 
 use v5.14;
 use warnings;
@@ -19,8 +18,6 @@ C<Future::IO::System> - C<system()>-like methods for L<Future::IO>
 
 =head1 SYNOPSIS
 
-=for highlighter language=perl
-
    use Future::IO;
    use Future::IO::System;
 
@@ -143,17 +140,17 @@ sub run
       if( $want_in ) {
          close $infh[0];
          push @f, Future::IO->syswrite_exactly( $infh[1], $params{in} )
-            ->then( sub { Future->done() } );
+            ->then( sub { close $infh[1]; Future->done() } );
       }
 
       if( $want_out ) {
          close $outfh[1];
-         push @f, Future::IO->read_until_eof( $outfh[0] );
+         push @f, Future::IO->sysread_until_eof( $outfh[0] );
       }
 
       if( $want_err ) {
          close $errfh[1];
-         push @f, Future::IO->read_until_eof( $errfh[0] );
+         push @f, Future::IO->sysread_until_eof( $errfh[0] );
       }
 
       return Future->needs_all( @f );
